# Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy
import scipy

import Heuristic_stream
import MisraGries

asymHammingDistanceWeight = 0.1

def distanceToClosestCenter(centers, p):
	if len(centers) == 0:
		return numpy.Inf, 0

	minDistance = numpy.Inf

	for i in range(len(centers)):
		dist = distanceOfPoints(centers[i], p)

		if dist < minDistance:
			minDistance = dist
			closestCenterIndex = i

	return minDistance, closestCenterIndex

def distanceOfPoints(center, point):
	dist = 0
	i = 0
	for j in range( len(point) ):
		while i < len(center) and center[i] < point[j]:
			i += 1
			dist += asymHammingDistanceWeight # center[i] is center but not in point

		' we finished processing center and all entries left in point must increase the distance '
		if i >= len(center):
			dist += len(point) - j # the entries are in point but not in center
			break

		if center[i] != point[j]:
			dist += 1 # point[j] is in point but not in center
		else:
			i += 1

	if i < len(center):
		dist += asymHammingDistanceWeight*(len(center) - i) # the entries are in center but not in point
	
	return dist

def recoverRightClusters(finalMGs, minClusterSize, threshold='auto', cols=None):
	print('len(finalMGs): {}'.format(len(finalMGs)))
	rightClusters = []
	for i in range( len(finalMGs) ):
		MG = finalMGs[i]
		transactionCount = MG.numVertices

		if transactionCount < minClusterSize:
			continue

		cluster = set()
		if threshold == 'auto':
			ps = numpy.arange(0.5,1,0.1)
			qs = numpy.arange(0.01,0.11,0.01)
			_,_,cluster = Heuristic_stream.estimatePQ(ps,qs,MG,cols)
		else:
			cluster = recoverRightCluster(MG, threshold)

		if len(cluster) > 0:
			rightClusters.append(cluster)

	return rightClusters

def recoverRightCluster(MG, threshold):
	cluster = set()

	for key in MG.counters:
		count = MG.counters[key]
		if count / MG.numVertices >= threshold:
			cluster.add(key)

	return cluster

def convertLineToVector(line):
	return numpy.fromstring(line, sep=',')

def writeClusterMatrix(outfile, clusters, columns):
	''' Writes the clusters into a matrix as a csv file. '''
	rows = len(clusters)
	B = numpy.zeros((rows, columns), dtype=numpy.float64)
	i = 0
	for cluster in clusters:
		if len(cluster) == 0:
		   continue

		for item in cluster:
		   B[i,item] = 1
		i += 1
	
	B = B[0:i,:] # rescale B to make it smaller in case we found some empty cluster

	numpy.savetxt(outfile, B, delimiter=',')

''' Returns [] if the fp ended. '''
def getNextPointDenseInput(fp):
	line = fp.readline()
	if line:
		denseVector = convertLineToVector(line)
		return list( numpy.nonzero(denseVector)[0] ), len(denseVector)
	else:
		return [], 0

''' Returns [] if the fp ended. '''
def getNextPointSparseInput(fp):
	p = []

	line = fp.readline()
	if not line:
		return p,-1
	
	lineSplit = str.split(line, ',')
	left = int(lineSplit[0])
	right = int(lineSplit[1])

	isSorted = True
	lastRight = right
	while line:
		lineSplit = str.split(line, ',')
		curLeft = int(lineSplit[0])
		curRight = int(lineSplit[1])

		if left != curLeft:
			fp.seek(lastPos)
			break

		if curRight < lastRight:
			isSorted = False

		p.append(curRight)

		lastRight = curRight
		lastPos = fp.tell()
		line = fp.readline()

	if not isSorted:
		p.sort()

	return p,left

def readSparseMatrix(inputfile, skipHeader=False):
	data = []
	row_ind = []
	col_ind = []

	with open(inputfile) as fp:
		line = fp.readline()
		if skipHeader:
			line = fp.readline()

		while line:
			lineSplit = str.split(line, ',')
			row_ind.append(int(lineSplit[0]))
			col_ind.append(int(lineSplit[1]))
			data.append(1)
		
			line = fp.readline()

	return scipy.sparse.csr_matrix((data, (row_ind, col_ind)))

def writeReconstructedClustersToFile(outputfile, rightClusters, writeEmptyClusters=False):
	# writes clusters line by line to outputfile
	with open(outputfile,'w') as outFile:
		for p in rightClusters:
			if not writeEmptyClusters and len(p) == 0:
				continue

			cluster = list(p)
			for i in range(len(cluster)-1):
				item = cluster[i]
				outFile.write('{0},'.format(item))
			if len(cluster) > 0:
				outFile.write(f'{cluster[len(cluster)-1]}')
			outFile.write('\n')

