# Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy
import scipy

import Heuristic

def BinaryMajorityVote(B, transactionClusters, threshold=0.5, minClusterSize=1, useHeuristic=False):
	clusters = list()
	for transactionCluster in transactionClusters:
		if len(transactionCluster) < minClusterSize:
			continue

		Bsubmatrix = B[list(transactionCluster), :]
		cluster = BinaryMajorityVoteSingleClusterWithVotingScheme(Bsubmatrix, threshold)
		clusters.append(cluster)

	return clusters

def BinaryMajorityVoteSingleClusterWithVotingScheme(B, threshold):
	cluster = set()

	if threshold == 'auto':
		ps = numpy.arange(0.3,1,0.05)
		qs = numpy.arange(0.01,0.11,0.01)
		_,_,cluster = Heuristic.estimatePQ(ps, qs, B)
	else:
		cluster = BinaryMajorityVoteSingleCluster(B, threshold)
	
	return cluster

def BinaryMajorityVoteSingleCluster(B, threshold):
	'''
	Returns a set of all column indices where the fractions of 1s in the column
	is larger than threshold.
	'''
	columnIndices = set()

	rows,cols = B.shape

	nonzeroCountPerCol = []
	if isinstance(B,scipy.sparse.csr.csr_matrix):
		nonzeroCountPerCol = B.getnnz(0)
	else:
		nonzeroCountPerCol = numpy.count_nonzero(B, axis=0)

	absThreshold = rows * threshold

	for col in range(cols):
		if nonzeroCountPerCol[col] > absThreshold:
			columnIndices.add(col)

	return columnIndices

