# Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import copy
import numpy
from scipy.sparse import csr_matrix
from sklearn.cluster import KMeans

import HelperFunctions
import MisraGries
import OnlineFL

class PLS:
	def __init__(self,
				 k,
				 initialLowerBound,
				 maxNumCenters = 300,
				 beta = 2.0,
				 MGMaxCounters = 200,
				 clusteringMethod='kmeans',
				 s = 200):
		self.lowerBound = initialLowerBound
		self.k = k
		self.beta = beta
		self.MGMaxCounters = MGMaxCounters
		self.maxNumCenters = maxNumCenters # 30 * self.k # in the paper this is 4k(1 + log n)(1 + 4(gamma + beta)) 
		self.clusteringMethod = clusteringMethod
		self.s = s

		self.numPoints = 0
		self.numDiscardedPoints = 0
		self.phase = 1

		self.centers = []
		self.centerMGs = []

		self.initializeOnlineFL()

		' used as means of an output '
		self.finalMGs = []

	def addPoint(self, p):
		self.numPoints += 1

		nnz = len(p)
		if nnz < self.s: # only process points which obey the degree restriction
			self.processStreamPoint(p)
		else:
			self.numDiscardedPoints += 1
	
	def processStreamPoint(self, p):
		MG = MisraGries.MisraGries(self.MGMaxCounters)
		MG.addNumbersFromBinaryVector(p)
		success = self.onlineFL.addPoint( copy.deepcopy(p), MG )

		' start a new phase if we could not process point '
		if not success:
			self.startNewPhase()
			self.processStreamPoint(p) # we still need to process p

	def startNewPhase(self):
		print('Starting new phase after {} points.'.format(self.numPoints))
		self.phase += 1
		self.lowerBound = self.lowerBound * self.beta

		self.getCentersFromOnlineFL()
		self.initializeOnlineFL()

	def getCentersFromOnlineFL(self):
		self.centers = copy.deepcopy( self.onlineFL.centers )
		self.centerMGs = copy.deepcopy( self.onlineFL.MGs )


	''' Initializes the OnlineFL algorithm.
		The weighted centers from the previous phase are already
		added to the OnlineFL instance.
	'''
	def initializeOnlineFL(self):
		f = self.lowerBound / self.k # in the paper this is lowerBound / (k*(1 + lg n))
		maxSolutionCost = 10 * self.lowerBound # in the paper this is gamma*lowerBound where gamma is really large

		self.onlineFL = OnlineFL.OnlineFL(f, self.maxNumCenters, maxSolutionCost)
		self.onlineFL.addPointsFromCenters(self.centers, self.centerMGs)
	
	''' Choices for clustering methods are kmeans and kmedian '''
	def finishProcessingStream(self):
		print('Processed {} points of which {} were discarded because they had {} or more non-zero entries.'.format(self.numPoints, self.numDiscardedPoints, self.s, self.clusteringMethod))

		self.getCentersFromOnlineFL()
		centerWeights = self.vectorWithCenterWeights()
		print('Centers have weights: {}'.format(centerWeights))
	
		pointsToCenter = []
		if self.clusteringMethod == 'doNothing':
			self.finalMGs = self.centerMGs
			return
		elif self.clusteringMethod == 'kmeans':
			k = min(self.k, len(self.centers)) # necessary to avoid crashes of scikit-learn

			centerMatrix = self.getCenterMatrix()
			kmeans = KMeans(n_clusters=int(k))
			pointsToCenter = kmeans.fit_predict(centerMatrix, sample_weight=centerWeights)
		else:
			print('Unknown clustering method presented.')
			return

		self.computeFinalMGs(pointsToCenter)

	def vectorWithCenterWeights(self):
		centerWeights = []
		for i in range(len(self.centers)):
			centerWeights.append( self.centerMGs[i].numVertices )
		return centerWeights

	def getCenterMatrix(self):
		nnz = 0
		row_ind = []
		col_ind = []
		for i in range(len(self.centers)):
			center = self.centers[i]
			nnz_center = len(center)

			nnz += nnz_center
			row_ind.extend( [i] * nnz_center)
			col_ind.extend( center )
		
		centerMatrix = csr_matrix(([1] * nnz, (row_ind, col_ind)))

		return centerMatrix

	def computeFinalMGs(self, pointsToCenter):
		' initialize final MGs '
		self.finalMGs = []
		for i in range(self.k):
			self.finalMGs.append( MisraGries.MisraGries(self.MGMaxCounters) )

		' set the correct counts for the MGs of the final centers '
		onlineFLPoints = self.onlineFL.centers
		onlineFLMGs = self.onlineFL.MGs
		for i in range( len(onlineFLPoints) ):
			finalCenterIdx = pointsToCenter[i]
			finalCenterMG = self.finalMGs[finalCenterIdx]
			
			finalCenterMG.addCounters(onlineFLMGs[i])

			onlineFLMGs[i] = 0

		' clean up the counters of the MGs of the final centers '
		for i in range(self.k):
			self.finalMGs[i].removeSuperfluousCounters()

