# Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy
import random
import time

import CoverLeftSide
import HelperFunctions
import PLS

'''
	The main function of the sofa algorithm from the following publication:
		Biclustering and Boolean Matrix Factorization in Data Streams
		by Stefan Neumann and Pauli Miettinen
		PVLDB, 13(10): 1709-1722, 2020
		DOI: https://doi.org/10.14778/3401960.3401968
	If you use this publication in one of your papers, you need to cite the above
	publication. The source code is licensed under an XXX license.

	RETURNS: rightClusters, leftClusters, rows, cols
		rightClusters:
			A dict containing a list of right-side clusters for each threshold
			in thresholds. The vertices of each cluster are stored in a set.
		leftClusters:
			A dict containing a list of left-side clusters for each threshold
			in thresholds. The vertices of each cluster are stored in a list.
			leftClusters is only computed if coverLeftSide is not None.
		rows: The number of rows of the input matrix (=#left-side vertices of the graph).
		cols: The number of columns of the input matrix (=#right-side vertices of the graph).

	To understand the parameters that are required by sofa, see the end of
	main.py where the command-line options are described. The names are
	essentially identical.
'''
def sofa(filepath,
		 k,
		 thresholds=['auto'],
		 maxNumCenters=200,
		 beta=2,
		 MGMaxCounters=200,
		 minClusterSize=10, 
		 clusteringMethod='kmeans', 
		 s=200,
		 asymHammingDistanceWeight=0.1,
		 fileIsSparse=False,
		 skipHeader=False,
		 coverLeftSide=None,
		 minOverlap=0.0):
	print('Running sofa with the following parameters:')
	print(f'	thresholds: {thresholds}')
	print(f'	maxNumCenters: {maxNumCenters}')
	print(f'	MGMaxCounters: {MGMaxCounters}')
	print(f'	s: {s}')
	print(f'	k: {k}')
	print(f'	clustering: {clusteringMethod}')
	print(f'	coverLeftSide: {coverLeftSide}')
	print(f'	beta: {beta}')
	print(f'	asymHammingDistanceWeight: {asymHammingDistanceWeight}')
	print(f'	minClusterSize: {minClusterSize}')
	print(f'	fileIsSparse: {fileIsSparse}')
	print(f'	skipHeader: {skipHeader}\n')

	' we start by recovering the right clusters '
	start = time.time()
	rightClusters, rows, cols = sofaRightClusters(
		filepath,
		k,
		thresholds=thresholds,
		maxNumCenters=maxNumCenters,
		beta=beta,
		MGMaxCounters=MGMaxCounters,
		minClusterSize=minClusterSize,
		clusteringMethod=clusteringMethod, 
		s=s,
		asymHammingDistanceWeight=asymHammingDistanceWeight,
		fileIsSparse=fileIsSparse,
		skipHeader=skipHeader)
	end = time.time()
	print(f'Recovering right side clusters took {end-start} seconds.')

	' compute left side clusters '
	leftClusters = dict()
	if coverLeftSide != None:
		start = time.time()

		for threshold in thresholds:
			clustersForThreshold = rightClusters[threshold]

			''' since we are covering the left side we want to prune empty
				clusters and we have to put this information back into
				rightClusters dict to ensure that numbering of right and
				left clusters is consistent '''
			clustersForThreshold = [p for p in clustersForThreshold if len(p) > 0]
			rightClusters[threshold] = clustersForThreshold
			
			rightClusters[threshold], leftClusters[threshold] = CoverLeftSide.coverLeftSide(filepath,
					clustersForThreshold, coveringType=coverLeftSide,
					minOverlap=minOverlap, fileIsSparse=fileIsSparse,
					skipHeader=skipHeader, k=k)

			end = time.time()
			print(f'Finished covering data for threshold {threshold}; took {end-start} seconds since starting covering.')

		end = time.time()
		print(f'Recovering left side clusters took {end-start} seconds.')

	return rightClusters,leftClusters,rows,cols

def sofaRightClusters(filepath,
		 k,
		 thresholds=['auto'],
		 maxNumCenters=200,
		 beta=2,
		 MGMaxCounters=200,
		 minClusterSize=10, 
		 clusteringMethod='kmeans', 
		 s=200,
		 asymHammingDistanceWeight=0.1,
		 fileIsSparse=False,
		 skipHeader=False):
	HelperFunctions.asymHammingDistanceWeight = asymHammingDistanceWeight

	' compute right side clusters '
	lowerBound = 1

	pls = PLS.PLS(k,
				  lowerBound,
				  maxNumCenters=maxNumCenters,
				  beta=beta,
				  MGMaxCounters=MGMaxCounters,
				  clusteringMethod=clusteringMethod,
				  s=s)

	rows = 0
	cols = 0
	with open(filepath) as fp:
		if fileIsSparse and skipHeader:
			fp.readline()

		while True:
			if fileIsSparse:
				p,_ = HelperFunctions.getNextPointSparseInput(fp)
			else:
				p, curCols = HelperFunctions.getNextPointDenseInput(fp)
			rows += 1

			if len(p) == 0:
				break

			pls.addPoint(p)

			if fileIsSparse:
				cols = max(cols, max(p)+1)
			else:
				cols = max(cols,curCols)

	pls.finishProcessingStream()

	rightClusters = dict()
	for threshold in thresholds:
		rightClusters[threshold] = HelperFunctions.recoverRightClusters(pls.finalMGs, minClusterSize, threshold=threshold, cols=cols)

	return rightClusters, rows, cols

